//FILE:         degradation.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include "degradation.h"
#include "devil.h"
#include "ppm.h"
#include "gl_utils.h"

///////////////////////////////////////////////////////////////////////////////
gl_texture
load_rgb_image( const std::string& file )
{
#if HAVE_DEVIL
   return devil_load_rgb_gl( file );
#else
   return ppm_load_rgb_gl( file );
#endif
}



///////////////////////////////////////////////////////////////////////////////
grey_image*
load_grey_image( const std::string& file )
{
#if HAVE_DEVIL
   return devil_load_grey( file );
#else
   return ppm_load_grey( file );
#endif
}



///////////////////////////////////////////////////////////////////////////////
static void
copy_channel( rgba_image *dst, grey_image *src, unsigned int channel )
{
   grey_image::pixel_t pixel;
   for( unsigned int row = 0; row < dst->h; ++row ) {
      for( unsigned int col = 0; col < dst->w; ++col ) {
         src->get( row, col, pixel );
         dst->set( row, col, channel, pixel[ 0 ] );
      }
   }
}



///////////////////////////////////////////////////////////////////////////////
static bool
copy_image_to_channel( rgba_image *dst, const std::string& filename,
   unsigned int channel )
{
   if( filename.empty( ) ) {
      //user didn't supply a map to use.  use the default white
      return true;
   }

   //load the map
   grey_image *chan_img = load_grey_image( filename );
   if( !chan_img ) {
      std::cerr << "error: could not load map: " << filename << std::endl;
      return false;
   }

   //make sure the dimensions of the images match
   if( (dst->w != chan_img->w) || (dst->h != chan_img->h) ) {
      std::cerr << "error: " << filename 
         << ": dimensions of all channels must match" << std::endl;
      return false;
   }

   //copy the map to the channel
   copy_channel( dst, chan_img, channel );
   delete chan_img;

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
load_degradation_image( const cmd_args& args )
{
   //locals
   rgba_image rgba;
   gl_texture texture;
   rgba_image::pixel_t white = { 255, 255, 255, 255 };

   //load the lod map first
   grey_image *chan_img = load_grey_image( args.lod_map );
   if( !chan_img ) {
      std::cerr << "error: could not load lod map: " << args.lod_map << std::endl;
      return texture;
   }

   //now that we know the dimensions of the degradation map, alloc mem for it
   rgba.resize( chan_img->h, chan_img->w );
   rgba.fill( white );

   //copy the lod map to the alpha channel
   copy_channel( &rgba, chan_img, 3 );
   delete chan_img;

   if( !copy_image_to_channel( &rgba, args.red_map, 0 ) ) {
      std::cerr << "error: could not load channel map: " << args.red_map << std::endl;
      return texture;
   }

   if( !copy_image_to_channel( &rgba, args.green_map, 1 ) ) {
      std::cerr << "error: could not load channel map: " << args.green_map << std::endl;
      return texture;
   }

   if( !copy_image_to_channel( &rgba, args.blue_map, 2 ) ) {
      std::cerr << "error: could not load channel map: " << args.blue_map << std::endl;
      return texture;
   }

   //success
   rgba.flip_horizontal( );
   texture = gl_load( &rgba );
   return texture;
}
