//FILE:         gcd_glut.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include <unistd.h>
#include <memory>
#include "extgl.h"
#include "args.h"
#include "gl_utils.h"
#include "gl_fragment_program.h"
#include "gl_vertex_program.h"
#include "gaze_backend.h"
#include "gaze_driver.h"
#include "gaze_view.h"
#include "ppm.h"
#include "degradation.h"
#include <GL/glut.h>

//file local structure so glut callbacks can communicate
struct glut_t
{
   cmd_args args;
   std::auto_ptr<gcd_view> gcd;
   std::auto_ptr<gaze_driver> gaze;
   rgb_image screenshot;

};
static glut_t glut;

//these variables are extern'd to gaze_glut.cpp.  this is because
//there is no way to query the mouse state in glut.
int glut_mouse_x;
int glut_mouse_y;



///////////////////////////////////////////////////////////////////////////////
void glut_idle_cb( void )
{
   if( glut.gaze->need_update( ) ) {
      glutPostRedisplay( );
   } else {
      usleep( 700 );
   }
}



///////////////////////////////////////////////////////////////////////////////
void 
glut_motion_cb( int x, int y )
{
   glut_mouse_x = x;
   glut_mouse_y = y;
   glutPostRedisplay( );
}



///////////////////////////////////////////////////////////////////////////////
void 
glut_keyboard_cb( unsigned char key, int x, int y )
{
   switch( key ) {
      case 27:
         //escape
         exit( 0 );
      break;

      case 'p':
         //take a screenshot (write to screenshot.ppm")
         glReadPixels( 0, 0, glut.args.w, glut.args.h, GL_RGB, 
            GL_UNSIGNED_BYTE, glut.screenshot.pixels );
         glut.screenshot.flip_horizontal( );
         ppm_ascii_write( "screenshot.ppm", &glut.screenshot );
         std::cout << "wrote: screenshot.ppm" << std::endl;
      break;

      case '1':
         glut.gcd->set_mipmapping( GL_LINEAR_MIPMAP_LINEAR );
         std::cout << "mipmapping: GL_LINEAR_MIPMAP_LINEAR" << std::endl;
         glutPostRedisplay( );
      break;

      case '2':
         glut.gcd->set_mipmapping( GL_NEAREST_MIPMAP_LINEAR );
         std::cout << "mipmapping: GL_NEAREST_MIPMAP_LINEAR" << std::endl;
         glutPostRedisplay( );
      break;

      case '3':
         glut.gcd->set_mipmapping( GL_LINEAR_MIPMAP_NEAREST );
         std::cout << "mipmapping: GL_LINEAR_MIPMAP_NEAREST" << std::endl;
         glutPostRedisplay( );
      break;

      case '4':
         glut.gcd->set_mipmapping( GL_NEAREST_MIPMAP_NEAREST );
         std::cout << "mipmapping: GL_NEAREST_MIPMAP_NEAREST" << std::endl;
         glutPostRedisplay( );
      break;

      default:
         //std::cerr << "char: " << (char)key << " int: " << (int) key << std::endl;
      break;
   };
}



///////////////////////////////////////////////////////////////////////////////
void 
glut_display_cb( void )
{
   //clear color buffer
   glClear( GL_COLOR_BUFFER_BIT );

   //get then set the gaze position
   int mx, my;
   glut.gaze->get_gaze( mx, my );
   glut.gcd->set_eye( mx, my );

   //draw the image/gaze
   glut.gcd->render( );

   //swap buffers
   glutSwapBuffers( );
}



///////////////////////////////////////////////////////////////////////////////
bool
glut_init( int argc, char *argv[ ] )
{
   glutInit( &argc, argv );

   //we don't need the depth buffer
   glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA );

   //set window size
   glutInitWindowSize( glut.args.w, glut.args.h );

   //open a window
   glutCreateWindow( "Gaze Contingent Display" );

   //set event callbacks
   glutIdleFunc( glut_idle_cb );
   glutPassiveMotionFunc( glut_motion_cb );
   glutMotionFunc( glut_motion_cb );
   glutKeyboardFunc( glut_keyboard_cb );
   glutDisplayFunc( glut_display_cb );

   //alloc mem for taking screenshots
   glut.screenshot.resize( glut.args.h, glut.args.w );

   //init gl
   if( !gl_init( glut.args ) ) {
      std::cerr << gl_get_error( ) << std::endl;
      return false;
   }

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
int 
main( int argc, char *argv[ ] )
{
   //command line fun
   if( !glut.args.parse_args( argc, (const char**)argv ) ) return 1;

   //start up glut/gl
   if( !glut_init( argc, argv ) ) return 1;

   //read in the image
   gl_texture img_texture = load_rgb_image( glut.args.img_map );
   if( img_texture.error( ) ) {
      std::cerr << "error: could not open: " << glut.args.img_map << std::endl;
      return 1;
   }

   //load the degradation textures
   gl_texture lod_texture = load_degradation_image( glut.args );
   if( lod_texture.error( ) ) {
      std::cerr << "error: could not open: " << glut.args.lod_map << std::endl;
      return 1;
   }

   //lod texture should have a 1:1 aspect ratio
   if( lod_texture.get_width( ) != lod_texture.get_height( ) ) {
      std::cerr << "error: " << glut.args.lod_map << ": must be square" << std::endl;
      return 1;
   }

   //load up the gaze input driver
   glut.gaze.reset( init_gaze( glut.args ) );
   if( !glut.gaze.get( ) ) {
      std::cerr << "error: could not initialize gaze driver" << std::endl;
      return 1;
   }

   //load fragment program
   gl_fragment_program fp( gl_program::load( glut.args.fragment ) );
   if( fp.error( ) ) {
      std::cerr << "error: could not load fragment program" << std::endl;
      return 1;
   }

   //load vertex program
   gl_vertex_program vp( gl_program::load( glut.args.vertex ) );
   if( vp.error( ) ) { 
      std::cerr << "error: could not load vertex program" << std::endl;
      return 1;
   }

   //create a viewer object
   glut.gcd.reset( new gcd_view( img_texture, lod_texture, fp, vp, glut.args ) );
   glut.gcd->set_mipmapping( glut.args.mipmap.mipmap );

   //let the magic begin
   glutMainLoop( );

   //no errors
   return 0;
}
