//FILE:         gcd_sdl.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include <unistd.h>
#include "extgl.h"
#include "SDL.h"
#include "args.h"
#include "gl_utils.h"
#include "gl_fragment_program.h"
#include "gl_vertex_program.h"
#include "gaze_backend.h"
#include "gaze_driver.h"
#include "gaze_view.h"
#include "ppm.h"
#include "degradation.h"
#include "xine_dump.h"

class frame_locker: public xine_dump_callback
{
 public:
   frame_locker( gcd_view *gcd ): m_gcd( gcd ), 
      m_width( 0 ), m_height( 0 ), m_pixels( false ),
      m_update( 0 ) {
      pthread_mutex_init( &m_mutex, 0 );
   }

   virtual ~frame_locker( void ) {
      pthread_mutex_destroy( &m_mutex );
   }

   virtual void operator()( unsigned int w, unsigned int h, 
                            unsigned char *pixels )
   {
      lock( );
      m_width = w;
      m_height = h;
      if( !m_pixels ) {
         m_pixels = new unsigned char[ w * h * 3 ];
      }
      memcpy( m_pixels, pixels, w * h * 3 );
      ++m_update;
      unlock( );
   }

   bool need_update( void ) {
      lock( );
      bool update = ((m_update > 0) ? true : false);
      unlock( );
      return update;
   }

   void update( void ) {
      //make sure another thread isn't accessing the texture
      lock( );

      //delete old texture (gl_texture handles deleting null texture)
      m_texture.erase( );

      //load new texture
      m_texture = gl_load( m_width, m_height, m_pixels );

      //tell gcd about new texture
      m_gcd->set_texture( m_texture );

      m_update = 0;
      /*
      --m_update;
      if( m_update < 0 ) {
         m_update = 0;
      }
      */

      //done
      unlock( );
   }

   void lock( void ) {
      pthread_mutex_lock( &m_mutex );
   }

   void unlock( void ) {
      pthread_mutex_unlock( &m_mutex );
   }

 private:
   gcd_view *m_gcd;
   gl_texture m_texture;
   pthread_mutex_t m_mutex;

   unsigned int m_width, m_height;
   unsigned char *m_pixels;
   int m_update;
};



///////////////////////////////////////////////////////////////////////////////
bool
sdl_init( const cmd_args& args )
{
   //initialize sdl
   if( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
      std::cerr << "error: " << SDL_GetError( ) << std::endl;
      return false;
   }
   atexit( SDL_Quit );

   //we can add a fullscreen flag here if we want
   int flags = SDL_OPENGL;
   if( args.fullscreen ) {
      flags |= SDL_FULLSCREEN;
   }

   //initialize video
   SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
   SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
   SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
   SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8 );
   SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );
   SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
   SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
   SDL_Surface *screen = SDL_SetVideoMode( args.w, args.h, 32, flags );
   if( 0 == screen ) {
      std::cerr << "error: " << SDL_GetError( ) << std::endl;
      return false;
   }

   //init gl
   if( !gl_init( args ) ) {
      std::cerr << gl_get_error( ) << std::endl;
      return false;
   }

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
int 
main( int argc, char *argv[ ] )
{
   //command line fun
   cmd_args args;
   if( !args.parse_args( argc, (const char**)argv ) ) return 1;

   //start up sdl/gl
   if( !sdl_init( args ) ) return 1;

   //open video
   xine_dump frame_dumper;
   if( !frame_dumper.open( args.img_map ) ) {
      std::cerr << "error: could not open: " << args.img_map
         << ": " << frame_dumper.get_error( ) << std::endl;
      return 1;
   }

   /*
   //read in the image
   gl_texture img_texture = load_rgb_image( args.img_map );
   if( img_texture.error( ) ) {
      std::cerr << "error: could not open: " << args.img_map << std::endl;
      return 1;
   }
   */

   //load the degradation textures
   gl_texture lod_texture = load_degradation_image( args );
   if( lod_texture.error( ) ) {
      std::cerr << "error: could not open: " << args.lod_map << std::endl;
      return 1;
   }

   //lod texture should have a 1:1 aspect ratio
   if( lod_texture.get_width( ) != lod_texture.get_height( ) ) {
      std::cerr << "error: " << args.lod_map << ": must be square" << std::endl;
      return 1;
   }

   //load up the gaze input driver
   gaze_driver *gaze = init_gaze( args );
   if( !gaze ) {
      std::cerr << "error: could not initialize gaze driver" << std::endl;
      return 1;
   }

   //load fragment program
   gl_fragment_program fp( gl_program::load( args.fragment ) );
   if( fp.error( ) ) {
      std::cerr << "error: could not load fragment program" << std::endl;
      return 1;
   }

   //load vertex program
   gl_vertex_program vp( gl_program::load( args.vertex ) );
   if( vp.error( ) ) { 
      std::cerr << "error: could not load vertex program" << std::endl;
      return 1;
   }

   //create a viewer object
   gl_texture tmp;
   gcd_view gcd( tmp, lod_texture, fp, vp, args );
   gcd.set_mipmapping( args.mipmap.mipmap );
   gcd.set_flip( true );

   //create a lock object which will ensure that textures are safely
   //uploaded to opengl
   frame_locker lock( &gcd );
   frame_dumper.set_callback( &lock );

   //hide the cursor
   SDL_ShowCursor( false );

   //set window title
   SDL_WM_SetCaption( "Gaze Contingent Display", "Gaze Contingent Display" );

   //alloc mem for taking screenshots
   rgb_image screenshot;
   screenshot.resize( args.h, args.w );

   //start the video playing
   if( !frame_dumper.play( ) ) {
      std::cerr << "error: play: " << frame_dumper.get_error( ) << std::endl;
      return 1;
   }

   //timing vars
   Uint32 start_time, render_time;
   Uint32 time_sum = 0;
   Uint32 frames_rendered = 0;

   //main loop
   SDL_Event event;
   bool quit = false;
   bool update = false;
   int mx, my;
   while( !quit ) {
      if( SDL_PollEvent( &event ) ) {
         switch( event.type) {
            case SDL_QUIT:
               //close window button hit
               quit = true;
               update = true;
            break;

            case SDL_KEYDOWN:
               switch( event.key.keysym.sym ) {
                  case SDLK_ESCAPE:
                     //escape pressed, quit!
                     quit = true;
                     update = true;
                  break;

                  case SDLK_F12:
                     //take a screenshot (write to screenshot.ppm")
                     glReadPixels( 0, 0, args.w, args.h, GL_RGB, 
                        GL_UNSIGNED_BYTE, screenshot.pixels );
                     screenshot.flip_horizontal( );
                     ppm_ascii_write( "screenshot.ppm", &screenshot );
                     std::cout << "wrote: screenshot.ppm" << std::endl;
                  break;

                  case SDLK_1:
                     gcd.set_mipmapping( GL_LINEAR_MIPMAP_LINEAR );
                     std::cout << "mipmapping: GL_LINEAR_MIPMAP_LINEAR" << std::endl;
                     update = true;
                  break;

                  case SDLK_2:
                     gcd.set_mipmapping( GL_NEAREST_MIPMAP_LINEAR );
                     std::cout << "mipmapping: GL_NEAREST_MIPMAP_LINEAR" << std::endl;
                     update = true;
                  break;

                  case SDLK_3:
                     gcd.set_mipmapping( GL_LINEAR_MIPMAP_NEAREST );
                     std::cout << "mipmapping: GL_LINEAR_MIPMAP_NEAREST" << std::endl;
                     update = true;
                  break;

                  case SDLK_4:
                     gcd.set_mipmapping( GL_NEAREST_MIPMAP_NEAREST );
                     std::cout << "mipmapping: GL_NEAREST_MIPMAP_NEAREST" << std::endl;
                     update = true;
                  break;

                  case SDLK_LEFT:
                     frame_dumper.skip( -10 );
                     std::cout << "skip -10s" << std::endl;
                     update = true;
                  break;

                  case SDLK_RIGHT:
                     frame_dumper.skip( 10 );
                     std::cout << "skip 10s" << std::endl;
                     update = true;
                  break;

                  case SDLK_DOWN:
                     frame_dumper.skip( -300 );
                     std::cout << "skip -5min" << std::endl;
                     update = true;
                  break;

                  case SDLK_UP:
                     frame_dumper.skip( 300 );
                     std::cout << "skip 5min" << std::endl;
                     update = true;
                  break;

                  default: break;
               }
            break;

            case SDL_MOUSEMOTION:
               update = true;
            break;

            case SDL_VIDEOEXPOSE:
               update = true;
            break;

            default: break;
         }
      }

      //check if the gaze driver wants to update the screen
      if( gaze->need_update( ) ) {
          update = true;
      }

      if( lock.need_update( ) ) {
         update = true;
      }

      //do we need to update
      if( update ) {
         //record render start time
         start_time = SDL_GetTicks( );

         //clear color buffer
         glClear( GL_COLOR_BUFFER_BIT );

         //get then set the gaze position
         gaze->get_gaze( mx, my );
         gcd.set_eye( mx, my );

         if( lock.need_update( ) ) {
            lock.update( );
         }

         //draw the image/gaze
         gcd.render( );

         //no more updates
         update = false;

         //swap buffers
         SDL_GL_SwapBuffers( );

         //determine how long it took to render the frame
         render_time = SDL_GetTicks( ) - start_time;
         time_sum += render_time;
         ++frames_rendered;
      } else {
         //we don't need to update, give up the CPU
         usleep( 700 );
      }
   }

   //output average rendertime
   float avg = static_cast<float>( time_sum ) / frames_rendered;
   std::cerr << "avg render time: " << avg << "ms (" << 1000.0 / avg 
      << " FPS)" << std::endl;

   //remove gaze viewer object
   delete gaze;

   //no errors
   return 0;
}
