//FILE:         gl_program.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include <fstream>
#include "gl_program.h"

///////////////////////////////////////////////////////////////////////////////
gl_program::~gl_program( void )
{
   erase( );
}

///////////////////////////////////////////////////////////////////////////////
bool
gl_program::enable_bind( void )
{
   if( !m_handle ) {
      //program has not been compiled, do so now
      if( !compile( ) ) return false;
   }

   //bind the program
   glEnable( m_program_type );
   glBindProgramARB( m_program_type, m_handle );

   //upload local parameters
   for( param_list::const_iterator i = m_locals.begin( ); 
      i != m_locals.end( ); ++i ) {
      glProgramLocalParameter4fvARB( m_program_type, i->first, i->second );
   }

   m_is_bound = true;
   return true;
}

///////////////////////////////////////////////////////////////////////////////
bool
gl_program::disable( void )
{
   m_is_bound = false;
   glDisable( m_program_type );
   return true;
}

///////////////////////////////////////////////////////////////////////////////
bool
gl_program::compile( void )
{
   //make sure there is some code present
   if( m_code.empty( ) ) {
      std::cerr << "error: could not compile zero length code" << std::endl;
      return false;
   }
   
   //generate and bind a program handle (just like with textures)
   glGenProgramsARB( 1, &m_handle );
   glBindProgramARB( m_program_type, m_handle );

   //upload the program
   glProgramStringARB( m_program_type, GL_PROGRAM_FORMAT_ASCII_ARB,
      m_code.size( ), m_code.c_str( ) );

   //check for compile errors
   if( GL_INVALID_OPERATION == glGetError( ) ) {
      //find the error position
      GLint pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos );

      //print error message
      const GLubyte *error = glGetString( GL_PROGRAM_ERROR_STRING_ARB );
      std::cerr << "error: compile error: " << pos << ": " << error << std::endl;

      //return handle and return error
      erase( );
      return false;
   }

   //success
   return true;
}

///////////////////////////////////////////////////////////////////////////////
void
gl_program::set_local_parameter( unsigned int index, real_t s )
{
   m_locals[ index ].set( s, s, s, s );
   upload_local_param( index );
}

///////////////////////////////////////////////////////////////////////////////
void
gl_program::set_local_parameter( unsigned int index, real_t x, real_t y,
   real_t z, real_t w )
{
   m_locals[ index ].set( x, y, z, w );
   upload_local_param( index );
}

///////////////////////////////////////////////////////////////////////////////
void
gl_program::set_local_parameter( unsigned int index, 
   const vector& v, real_t w )
{
   m_locals[ index ].set( v.x( ), v.y( ), v.z( ), w );
   upload_local_param( index );
}

///////////////////////////////////////////////////////////////////////////////
void
gl_program::erase( void )
{
   if( m_handle != 0 ) {
      glDeleteProgramsARB( 1, &m_handle );
   }
}

///////////////////////////////////////////////////////////////////////////////
gl_program::gl_program( GLenum program_type, const std::string& code, 
   bool should_compile ):
   m_program_type( program_type ),
   m_handle( 0 ),
   m_code( code ),
   m_is_bound( false ),
   m_error( false )
{
   if( should_compile ) {
      //compile sets error state
      m_error = !compile( );
   }
}

///////////////////////////////////////////////////////////////////////////////
std::string
gl_program::load( const std::string& filename )
{
   std::ifstream handle( filename.c_str( ) );
   if( !handle ) return std::string( );
   return gl_program::load( handle );
}

///////////////////////////////////////////////////////////////////////////////
std::string
gl_program::load( std::istream& input )
{
   //load the program from disk (check out this stl madness!)
   std::istreambuf_iterator<char> begin( input ), end;
   std::string data( begin, end );
   return data;
}

///////////////////////////////////////////////////////////////////////////////
//will only upload param if this program is currently bound
void
gl_program::upload_local_param( unsigned int param )
{
   if( m_is_bound ) {
      param_list::const_iterator i = m_locals.find( param );
      glProgramLocalParameter4fvARB( m_program_type, i->first, i->second );
   }
}
