//FILE:         gl_utils.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include <cstdlib>
#include "extgl.h"
#include <GL/glu.h>
#include "gl_utils.h"

static const GLenum s_target = GL_TEXTURE_2D;
static std::string gl_error = "none";

///////////////////////////////////////////////////////////////////////////////
std::string
gl_get_error( void ) 
{
   return gl_error;
}



///////////////////////////////////////////////////////////////////////////////
bool
gl_init( const cmd_args& args )
{
   //check extensions
   extgl_Initialize( );

   //make sure the extensions we need are present
   if( !(extgl_ExtensionSupported( "GL_EXT_texture_rectangle" ) || 
         extgl_ExtensionSupported( "GL_NV_texture_rectangle" )) ) {
      gl_error = "error: texture rectangle extension is required";
      return false;
   }

   //we need fragment programs
   if( !extgl_ExtensionSupported( "GL_ARB_fragment_program" ) ) {
      gl_error = "error: fragment program support is required";
      return false;
   }

   //we need vertex programs
   if( !extgl_ExtensionSupported( "GL_ARB_vertex_program" ) ) {
      gl_error = "error: vertex program support is required";
      return false;
   }

   //clear color is black
   glClearColor( 0.0, 0.0, 0.0, 0.0 );

   //resize the window
   gl_resize( args.w, args.h );

   //disable any tests we will not need
   glDisable( GL_DEPTH_TEST );
   glDisable( GL_CULL_FACE );

   //enable texturing
   glEnable( s_target );
   glColor4f( 1.0, 1.0, 1.0, 1.0 );

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
void
gl_resize( unsigned int w, unsigned int h )
{
   //set viewport
   glViewport( 0, 0, w, h );

   //set orthographic projection
   glMatrixMode( GL_PROJECTION );
   glLoadIdentity( );
   glOrtho( 0.0, w, 0.0, h, -1.0, 1.0 );

   //switch back to modelview matrix
   glMatrixMode( GL_MODELVIEW );
   glLoadIdentity( );
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
gl_load( rgba_image *img )
{
   if( !img ) return gl_texture( );

   GLuint handle;
   GLenum target = s_target;
   glGenTextures( 1, &handle );
   glBindTexture( target, handle );
   glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE ); //FIXME decal?
   glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
   glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
   glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
   glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
   glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
   glPixelStorei( GL_UNPACK_SWAP_BYTES, GL_FALSE) ;

   //upload the texture
   if( GL_TEXTURE_RECTANGLE_EXT == s_target ) {
      //can not build mipmaps with GL_TEXTURE_RECTANGLE_EXT
      glTexImage2D( target, 0,
         GL_RGBA,
         img->w, img->h,
         0, GL_RGBA,
         GL_UNSIGNED_BYTE,
         img->pixels
      );
   } else {
      gluBuild2DMipmaps( target, GL_RGBA, img->w, img->h, GL_RGBA, 
         GL_UNSIGNED_BYTE, img->pixels );
   }

   gl_texture tex;
   tex.set_handle( handle );
   tex.set_width( img->w );
   tex.set_height( img->h );
   tex.set_target( target );
   return tex;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
gl_load( rgb_image *img )
{
   if( !img ) return gl_texture( );

   GLuint handle;
   GLenum target = s_target;
   glGenTextures( 1, &handle );
   glBindTexture( target, handle );
   glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE ); //FIXME decal?
   glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
   glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
   glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
   glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
   glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
   glPixelStorei( GL_UNPACK_SWAP_BYTES, GL_FALSE) ;

   //upload the texture
   if( GL_TEXTURE_RECTANGLE_EXT == s_target ) {
      //can not build mipmaps with GL_TEXTURE_RECTANGLE_EXT
      glTexImage2D( target, 0,
         GL_RGBA,
         img->w, img->h,
         0, GL_RGB,
         GL_UNSIGNED_BYTE,
         img->pixels
      );
   } else {
      gluBuild2DMipmaps( target, GL_RGBA, img->w, img->h, GL_RGB, 
         GL_UNSIGNED_BYTE, img->pixels );
   }

   gl_texture tex;
   tex.set_handle( handle );
   tex.set_width( img->w );
   tex.set_height( img->h );
   tex.set_target( target );
   return tex;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
gl_load( unsigned int w, unsigned int h, unsigned char *pixels )
{
   if( !pixels ) return gl_texture( );

   GLuint handle;
   GLenum target = s_target;
   glGenTextures( 1, &handle );
   glBindTexture( target, handle );
   glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE ); //FIXME decal?
   glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
   glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
   glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
   glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
   glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
   glPixelStorei( GL_UNPACK_SWAP_BYTES, GL_FALSE) ;

   //upload the texture
   if( GL_TEXTURE_RECTANGLE_EXT == s_target ) {
      //can not build mipmaps with GL_TEXTURE_RECTANGLE_EXT
      glTexImage2D( target, 0,
         GL_RGBA,
         w, h,
         0, GL_RGB,
         GL_UNSIGNED_BYTE,
         pixels
      );
   } else {
      gluBuild2DMipmaps( target, GL_RGBA, w, h, GL_RGB, 
         GL_UNSIGNED_BYTE, pixels );
   }

   gl_texture tex;
   tex.set_handle( handle );
   tex.set_width( w );
   tex.set_height( h );
   tex.set_target( target );
   return tex;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
gl_load( grey_image *img )
{
   if( !img ) return gl_texture( );

   GLuint handle;
   GLenum target = s_target;
   glGenTextures( 1, &handle );
   glBindTexture( target, handle );
   glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE ); //FIXME decal?
   glTexParameteri( target, GL_TEXTURE_WRAP_S, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_WRAP_T, GL_CLAMP );
   glTexParameteri( target, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
   glTexParameteri( target, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR );
   glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
   glPixelStorei( GL_UNPACK_SKIP_ROWS, 0 );
   glPixelStorei( GL_UNPACK_SKIP_PIXELS, 0 );
   glPixelStorei( GL_UNPACK_ROW_LENGTH, 0 );
   glPixelStorei( GL_UNPACK_SWAP_BYTES, GL_FALSE) ;

   //upload the texture
   if( GL_TEXTURE_RECTANGLE_EXT == s_target ) {
      //can not build mipmaps with GL_TEXTURE_RECTANGLE_EXT
      glTexImage2D( target, 0,
         GL_LUMINANCE,
         img->w, img->h,
         0, GL_LUMINANCE,
         GL_UNSIGNED_BYTE,
         img->pixels
      );
   } else {
      gluBuild2DMipmaps( target, GL_LUMINANCE, img->w, img->h, GL_LUMINANCE,
         GL_UNSIGNED_BYTE, img->pixels );
   }

   gl_texture tex;
   tex.set_handle( handle );
   tex.set_width( img->w );
   tex.set_height( img->h );
   tex.set_target( target );
   return tex;
}
