//FILE:         image.h
//AUTHOR:       Nathan Cournia <nathan@cournia.com>
//NOTES:        Partial specialization of template member functions is not allowed
//              by the C++ standard. :(  Some posts on comp.std.c++ seem to suggest
//              that it might be in the next draft though . . .  We can use partial
//              specialization to avoid the assumptions/branches in 
//              the get/set methods.  We can also optimize code for fixed vs.
//              float types.

#ifndef COURNIA_IMAGE_H
#define COURNIA_IMAGE_H 1

#include <iostream>
#include <iomanip>
#include <cassert>
#include <limits>

//#define DEBUG(x) (std::cerr << __FILE__ << ":" << __LINE__ << " " << __PRETTY_FUNCTION__ << ": " << x << std::endl)
#define DEBUG(x)

template <class T, unsigned int CHANNELS>
struct image
{
   T *pixels;              //tightly packed
   unsigned int w, h;      //cols, rows
   static const unsigned int channels;

   typedef T pixel_t[ CHANNELS ];

   image( void ): pixels( 0 ), w( 0 ), h( 0 ) { }

   ~image( void ) { erase( ); }

   void resize( unsigned int rows, unsigned int cols ) {
      if( w == cols && h == rows ) return;
      erase( );
      pixels = new T[ rows * cols * CHANNELS ];
      assert( pixels );
      w = cols; h = rows;
   }

   void erase( void ) {
      w = 0; h = 0;
      delete[] pixels;
      pixels = 0;
   }

   T get( unsigned int row, unsigned int col ) const {
      assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      return *p;
   }

   //warning: functions assumes 3 channels, no check is made for speed
   void get( unsigned int row, unsigned int col, T& r, T& g, T& b ) const {
      DEBUG( "w: " << w << " c: " << col );
      assert( CHANNELS > 2 ); assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      r = *p;
      g = *(p + 1);
      b = *(p + 2);
   }

   //warning: functions assumes 4 channels, no check is made for speed
   void get( unsigned int row, unsigned int col, T& r, T& g, T& b, T& a ) const {
      DEBUG( "w: " << w << " c: " << col );
      assert( CHANNELS > 3 ); assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      r = *p;
      g = *(p + 1);
      b = *(p + 2);
      a = *(p + 3);
   }

   void get( unsigned int row, unsigned int col, T color[ CHANNELS ] ) const {
      DEBUG( "w: " << w << " c: " << col );
      assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      for( unsigned int i = 0; i < CHANNELS; ++i ) {
         color[ i ] = *(p + i);
      }
   }

   T get( unsigned int row, unsigned int col, unsigned int chan ) {
      DEBUG( "w: " << w << " c: " << col );
      assert( chan < CHANNELS ); assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      return *(p + chan);
   }

   void set( unsigned int row, unsigned int col, T color[ CHANNELS ] ) {
      DEBUG( "w: " << w << " c: " << col );
      assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      for( unsigned int i = 0; i < CHANNELS; ++i ) {
         *(p + i) = color[ i ];
      }
   }

   void set( unsigned int row, unsigned int col, unsigned int chan, T color ) {
      DEBUG( "w: " << w << " c: " << col );
      assert( chan < CHANNELS ); assert( row < h ); assert( col < w ); assert( pixels );
      T *p = pixels + row * w * CHANNELS + col * CHANNELS;
      *(p + chan) = color;
   }

   void fill( T color[ CHANNELS ] ) {
      for( unsigned int r = 0; r < h; ++r ) {
         for( unsigned int c = 0; c < w; ++c ) {
            set( r, c, color );
         }
      }
   }

   void flip_horizontal( void ) {
      unsigned int row_size = w * CHANNELS;
      for( unsigned int r = 0; r < h / 2; ++r ) {
         for( unsigned int c = 0; c < w; ++c ) {
            for( unsigned int e = 0; e < CHANNELS; ++e ) {
               std::swap( 
                  *(pixels + r * row_size + c * CHANNELS + e),
                  *(pixels + (h - r - 1) * row_size + c * CHANNELS + e)
               );
            }
         }
      }
   }

   void normalize( void ) {
      //normalize
      T min = std::numeric_limits<T>::max( );
      T max = ((std::numeric_limits<T>::isinteger ) ?
         std::numeric_limits<T>::min : -std::numeric_limits<T>::max( ));
      T v;
      unsigned int size = w * h;

      for( unsigned int i = 0; i < size; ++i ) {
         v = pixels[ i * CHANNELS ];
         if( v < min ) min = v;
         if( v > max ) max = v;
      }
      T d = max - min;

      for( unsigned int i = 0; i < size; ++i ) {
         v = pixels[ i * CHANNELS ];
         v = (v - min) / d;
         pixels[ i * CHANNELS ] = v;
      }
   }
};

template<class T, unsigned int CHANNELS>
const unsigned int image<T, CHANNELS>::channels = CHANNELS;

typedef image<unsigned char, 4> rgba_image;
typedef image<unsigned char, 3> rgb_image;
typedef image<unsigned char, 1> grey_image;

#endif
