//FILE:         iscan.cpp
//AUTHOR:       Brian Cumming <brcummi@clemson.edu>
//              Angry Daniel <dcmccal@clemson.edu>
//              Nathan Cournia <nathan@cournia.com>

#include "iscan.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <cassert>

const iscan::byte iscan::PREFIX = 0x44;

const iscan::state iscan::RESET = 0x00;
const iscan::state iscan::CALIBRATE = 0x64;
const iscan::state iscan::RUN = 0xc8;

const unsigned int iscan::ERRORS = 5;

///////////////////////////////////////////////////////////////
///
///////////////////////////////////////////////////////////////
iscan::iscan( void ): m_device(-1), m_active(false), m_update_callback( 0 )
{
   create_thread( );

   lock_mutex( );

   m_eyes[ ISCAN_LEFT_EYE ].set_parent( this );
   m_eyes[ ISCAN_RIGHT_EYE ].set_parent( this );

   unlock_mutex( );
}



///////////////////////////////////////////////////////////////
/// \param device: the device to get eye tracker data from
/// \param m: the eye tracker mode
/// \param d: the dominance of the eyes
///////////////////////////////////////////////////////////////
iscan::iscan( const std::string & device, mode m, dominance d ): 
   m_device(-1), m_active(false), m_update_callback( 0 )
{
   create_thread( );

   lock_mutex( );

   m_eyes[ ISCAN_LEFT_EYE ].set_parent( this );
   m_eyes[ ISCAN_RIGHT_EYE ].set_parent( this );

   unlock_mutex( );

   open( device, m );
}



///////////////////////////////////////////////////////////////
///
///////////////////////////////////////////////////////////////
iscan::~iscan( void ) 
{
   close( );

   pthread_cancel( m_thread );

   pthread_mutex_destroy( &m_mutex );
}



///////////////////////////////////////////////////////////////
/// \param device: the device to get eye tracker data from
/// \param m: the eye tracker mode
/// \param d: the dominance of the eyes
/// \return - true if the device opened sucessfully
///         - false otherwise
///////////////////////////////////////////////////////////////
bool
iscan::open( const std::string & device, mode m, iscan::dominance d )
{
   assert( m_device < 0 );

   m_device = ::open( device.c_str( ), O_RDONLY );

   if( m_device < 0 ) {
      set_error( "could not open iscan device" );
      return false;
   }

   if( tcgetattr( m_device, &m_saved_settings ) < 0 ) {
      set_error( "could not save existing device settings" );
      return false;
   }

   termios settings;
   memset( &settings, 0, sizeof( termios ) );
   settings.c_cflag = CS8 | CLOCAL | CREAD | B115200 | CSTOPB;
   settings.c_iflag = IXOFF;
   settings.c_lflag = 0;
   settings.c_cc[ VMIN ] = 0;
   settings.c_cc[ VTIME ] = 20;

   clear_device( );

   if( tcsetattr( m_device, TCSANOW, &settings ) < 0 ) {
      set_error( "could not write new device settings" );
      return false;
   }

   clear_device( );

   lock_mutex( );

   m_num_eyes = m + 1;

   m_dominance = d;

   m_active = true;

   unlock_mutex( );

   return true;
}



///////////////////////////////////////////////////////////////
/// \return - true if the device closed sucessfully
///         - false otherwise
///////////////////////////////////////////////////////////////
bool
iscan::close( ) 
{
   if( m_device < 0 ) return true;

   //lock the device for the entire close operation, this ensures that 
   //the thread won't be reading from a closed device
   lock_mutex( );

   clear_device( );

   if( tcsetattr( m_device, TCSANOW, &m_saved_settings ) < 0 ) {
      unlock_mutex( );
      set_error( "could not restore saved device settings" );
      return false;
   }

   if( ::close( m_device ) < 0 ) {
      m_active = false;
      m_device = -1;
      unlock_mutex( );
      set_error( "could not close iscan device" );
      return false;
   }

   m_active = false;
   m_device = -1;
   unlock_mutex( );

   return true;
}



///////////////////////////////////////////////////////////////
/// \return - true if the device started sucessfully
///         - false otherwise
///////////////////////////////////////////////////////////////
bool
iscan::start( void )
{
   if( m_device < 0 ) {
      set_error( "iscan device not open" );
      return false;
   }

   lock_mutex( );
   m_active = true;
   unlock_mutex( );

   return true;
}



///////////////////////////////////////////////////////////////
/// \return - true if the device stopped sucessfully
///         - false otherwise
///////////////////////////////////////////////////////////////
bool
iscan::stop( void ) {
   if( m_device < 0 ) {
      set_error( "iscan device not open" );
      return false;
   }


   lock_mutex( );
   m_active = false;
   unlock_mutex( );

   return true;
}



///////////////////////////////////////////////////////////////
/// \param buffer: the buffer to read data into
/// \param size: the number of bytes to read
/// \return - true if the specified number of bytes were read
///         - false otherwise
///////////////////////////////////////////////////////////////
//FIXME cournia: looks like b/d copied this from dr. d's code.
//      the code doesn't always work.  bytes should be read in
//      one at a time.  also, the code should be made
//      interrupt safe by checking for EINTR.  look at
//      fob.cpp from http://fob.sf.net for an example
//      on how to read from a serial port safely.
bool
iscan::read( iscan::byte * buffer, unsigned int size )
{
   assert( m_active );

   unsigned int bytes_read = 0;
   unsigned int errors = 0;

   while( (bytes_read < size) && (errors < iscan::ERRORS) ) {
      int bytes = ::read( m_device, buffer + bytes_read, size - bytes_read );
      if( bytes < 0 ) {
         ++errors;
      } else {
         bytes_read += bytes;
      }
   }

   if( errors >= iscan::ERRORS ) {
      set_error( "iscan device read failure" );
      return false;
   }

   return true;
}



///////////////////////////////////////////////////////////////
/// \return - true if the update resulted in new data
///         - false otherwise
///////////////////////////////////////////////////////////////
bool
iscan::update( void )
{
   lock_mutex( );

   if( !m_active ) {
      unlock_mutex( );
      return false;
   }

   unlock_mutex( );

   iscan::byte buffer[ 8 ];

   while( buffer[ 0 ] != iscan::PREFIX ) {
      if( !read( &buffer[ 0 ], 1 ) ) return false;
   }

   if( !read( &buffer[ 1 ], 1 ) ) {
      if( buffer[ 1 ] != iscan::PREFIX ) return false;
   }

   if( !read( buffer, 8 ) ) return false;

   m_eyes[ ISCAN_LEFT_EYE ].extract( buffer );

   if( m_num_eyes == 2 ) {
      if( !read( buffer, 8 ) ) return false;

      m_eyes[ ISCAN_RIGHT_EYE ].extract( buffer );
   }

   switch( m_dominance ) {
      case iscan::LEFT_DOMINANCE: 
         m_eyes[ ISCAN_RIGHT_EYE ] = m_eyes[ ISCAN_LEFT_EYE ];
      break;

      case iscan::RIGHT_DOMINANCE: 
         m_eyes[ ISCAN_LEFT_EYE ] = m_eyes[ ISCAN_RIGHT_EYE ];
      break;

      case iscan::NO_DOMINANCE: break;
   }

   //if we've made it this far, we've got new data from the eye tracker
   //TODO cournia - only call the callback if the data has
   //               changed
   if( m_update_callback ) {
      //we don't lock here since the callback may call methods like
      //get_eye( ), which locks the mutex.  if locked here we
      //would get deadlock
      (*m_update_callback)( this );
   }

   return true;
}



///////////////////////////////////////////////////////////////
/// \return nothing
///////////////////////////////////////////////////////////////
void 
iscan::create_thread( void ) 
{
   pthread_mutex_init( &m_mutex, NULL );

   if( pthread_create( &m_thread, NULL, get_iscan_data, this ) != 0 ) {
      set_error( "could not create thread" );
   }
}



///////////////////////////////////////////////////////////////
///
///////////////////////////////////////////////////////////////
iscan::eye::eye( void ): 
   m_parent( NULL ),
   m_mapper( NULL ),
   m_responsible( false ),
   m_state( iscan::RESET ),
   m_x( 0.0 ),
   m_y( 0.0 ),
   m_diameter( 0.0 ) 
{
   //no-op
}



///////////////////////////////////////////////////////////////
///
///////////////////////////////////////////////////////////////
iscan::eye::~eye( void ) 
{
   if( m_responsible ) {
      delete m_mapper;
      m_mapper = NULL;
   }
}



///////////////////////////////////////////////////////////////
/// \param buffer: the buffer to extract the eye state from
/// \return nothing
///////////////////////////////////////////////////////////////
void
iscan::eye::extract( iscan::byte * buffer )
{
   m_parent->lock_mutex( );

   //TODO make endian safe
   m_state = (buffer[1] << 8) + buffer[0];
   m_x = (buffer[3] << 8) + buffer[2];
   m_y = (buffer[5] << 8) + buffer[4];
   m_diameter = (buffer[7] << 8) + buffer[6];

   m_parent->unlock_mutex( );
}



///////////////////////////////////////////////////////////////
/// \return the state of the eye in string form
///////////////////////////////////////////////////////////////
std::string
iscan::eye::get_state_string( void ) const 
{
   std::string s;
   get_state_string( s );
   return s;
}



///////////////////////////////////////////////////////////////
/// \param s: reference to eye state which will be filled
/// \return nothing
///////////////////////////////////////////////////////////////
void
iscan::eye::get_state_string( std::string & s ) const
{
   assert( m_parent );
   m_parent->lock_mutex( );

   switch( m_state ) {
      case RESET:
         s = "RESET";
      break;

      case CALIBRATE:
         s = "CALIBRATE";
      break;

      case RUN:
         s = "RUN";
      break;

      default:
         s = "INVALID STATE";
      break;
   }

   m_parent->unlock_mutex( );
}



///////////////////////////////////////////////////////////////
/// \param out: the stream to print to
/// \param eye: the eye to print
/// \return the resultant stream
///////////////////////////////////////////////////////////////
std::ostream & operator<<( std::ostream & out, const iscan::eye & eye ) {

   iscan::state state = eye.get_state( );

   out << "status:   ";

   switch (state) {
      case iscan::RESET:
         out << "RESET";
      break;

      case iscan::CALIBRATE:
         out << "CALIBRATE";
      break;

      case iscan::RUN:
         out << "RUN";
      break;

      default:
         out << "INVALID STATE";
      break;
   }

   out << std::endl;

   out << "x:        " << eye.get_x() << std::endl;
   out << "y:        " << eye.get_y() << std::endl;
   out << "diameter: " << eye.get_diameter() << std::endl;

   return out;
}

extern "C" 
{
   ///////////////////////////////////////////////////////////////
   /// \param data: pointer to the iscan object
   /// \return
   ///////////////////////////////////////////////////////////////
   void * get_iscan_data( void * data ) {

      pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );

      iscan * tracker = (iscan *)data;

      while (true) {
         tracker->update( );
      }
   }
} //end of extern "C"
