//FILE:         iscan.h
//AUTHOR:       Brian Cumming <brcummi@clemson.edu>
//              Angry Daniel <dcmccal@clemson.edu>
//              Nathan Cournia <nathan@cournia.com>

#ifndef UBER_ISCAN_H
#define UBER_ISCAN_H 1

#include <cassert>
#include <string>
#include <pthread.h>
#include <termios.h>
#include <unistd.h>
#include <iostream>
#include "error_base.h"
#include "mathematics.h"

// used to access an eye in monocular mode
static const unsigned int ISCAN_ONLY_EYE = 0;

// used to access eyes in binocular mode
static const unsigned int ISCAN_LEFT_EYE = 0;
static const unsigned int ISCAN_RIGHT_EYE = 1;

// the maximum number of eyes allowed
static const unsigned int ISCAN_MAX_EYES = 2;

extern "C" 
{
   /// \brief Function that is called by the iscan thread
   void * get_iscan_data(void * data);
}

///////////////////////////////////////////////////////////////
/// \brief Class to handle the access of eye tracker data
///////////////////////////////////////////////////////////////
class iscan : public error_base 
{
 public:
   typedef unsigned char byte;
   typedef unsigned int state;

   // two prefix bytes indicate the start of a new message
   static const byte PREFIX;

   // three possible states for each eye to be in
   static const state RESET;
   static const state CALIBRATE;
   static const state RUN;

   // maximum allowed read errors
   static const unsigned int ERRORS;

   // enum of possible eye tracker modes
   enum mode {
      MONOCULAR,
      BINOCULAR
   };

   // enum of potential eye dominance
   enum dominance {
      NO_DOMINANCE,
      LEFT_DOMINANCE,
      RIGHT_DOMINANCE
   };

   ///////////////////////////////////////////////////////////////
   /// \brief Class to define how the iscan data is to be mapped
   ///////////////////////////////////////////////////////////////
   class mapper 
   {
    public:
      /// \brief Destructor
      virtual ~mapper( void ) { }

      /// \brief Force implementation upon child classes
      virtual real_t map_x( real_t before ) const = 0;

      /// \brief Force implementation upon child classes
      virtual real_t map_y( real_t before ) const = 0;
   };

   ///////////////////////////////////////////////////////////////
   /// \brief Class to define a linear mapping of iscan data
   /// that results in values from 0.0 to 1.0
   ///////////////////////////////////////////////////////////////
   class mapper01: public mapper 
   {
    public:
      /// \brief Default and Conversion Constructor
      ///////////////////////////////////////////////////////////////
      /// \param max: the maximum value of the source data
      /// \warning mapping a value bigger than the max will result
      ///          in a value bigger than 1.0
      ///////////////////////////////////////////////////////////////
      mapper01( real_t x = 512.0, real_t y = 512.0 ) : m_x( x ), m_y( y ) {
         if( equals( m_x, 0.0 ) ) m_x = 1.0;
         if( equals( m_y, 0.0 ) ) m_y = 1.0;
      }

      /// \brief Destructor
      ~mapper01() { }

      /// \brief Apply the mapper
      ///////////////////////////////////////////////////////////////
      /// \param before: the source x value to convert
      /// \return the resultant x value after applying the map
      ///////////////////////////////////////////////////////////////
      real_t map_x( real_t before ) const { 
         return clamp( before / m_x, 0.0f, 1.0f ); 
      }

      /// \brief Apply the mapper
      ///////////////////////////////////////////////////////////////
      /// \param before: the source y value to convert
      /// \return the resultant y value after applying the map
      ///////////////////////////////////////////////////////////////
      real_t map_y( real_t before ) const { 
         return clamp( before / m_y, 0.0f, 1.0f ); 
      }

   private:
      /// The maximum x value of the source data
      real_t m_x;

      /// The maximum y value of the source data
      real_t m_y;
   }; 

   ///////////////////////////////////////////////////////////////
   /// \brief Class to encapsulate the state of an individual eye
   ///////////////////////////////////////////////////////////////

   class eye 
   {
    public:
      /// \brief Default Constructor
      eye( void );

      /// \brief Destructor
      ~eye( void );

      /// \brief Assignment Operator
      eye& operator=( const eye & e ) {
         m_parent = e.m_parent;
         m_mapper = e.m_mapper;
         m_responsible = false;
         m_state = e.m_state;
         m_x = e.m_x;
         m_y = e.m_y;
         m_diameter = e.m_diameter;
         return *this;
      }

      /// \brief Extract the eye state from the given buffer
      void extract( byte * buffer );

      /// \brief Set the iscan parent object
      ///////////////////////////////////////////////////////////////
      /// \param parent: the iscan parent object
      ///////////////////////////////////////////////////////////////
      inline void set_parent( iscan * parent ) { 
         m_parent = parent; 
      }

      /// \brief Set the mapper for the iscan data
      ///////////////////////////////////////////////////////////////
      /// \param map: the mapper to use when converting raw data 
      /// \param responsible: whether or not the class is responsible
      ///                     for the deletion of the mapper
      /// \warning be very careful with the responsible flag and
      ///          the memory management of mappers
      ///////////////////////////////////////////////////////////////
      inline void set_mapper( mapper * map, bool responsible = true ) {
         // we are reassigning the mapper, if one already exists
         // and we own the memory, delete it
         if( m_responsible ) {
            delete m_mapper;
         }

         // assign new mapper
         m_mapper = map;
         m_responsible = responsible;
      }

      /// \brief Get the state
      ///////////////////////////////////////////////////////////////
      /// \return the state of the eye
      ///////////////////////////////////////////////////////////////
      inline state get_state( void ) const {
         state s;
         get_state( s );
         return s;
      }

      /// \brief Get the state
      ///////////////////////////////////////////////////////////////
      /// \param s: reference to eye state which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_state( state & s ) const {
         assert( m_parent );
         m_parent->lock_mutex( );
         s = m_state;
         m_parent->unlock_mutex( );
      }

      /// \brief Get the state in string form
      std::string get_state_string( void ) const;

      /// \brief Get the state in string form
      void get_state_string( std::string & s ) const;

      /// \brief Get the x-coordinate
      ///////////////////////////////////////////////////////////////
      /// \return the mapped x-coordinate of the eye
      ///////////////////////////////////////////////////////////////
      inline real_t get_x( void ) const {
         real_t x;
         get_x( x );
         return x;
      }

      /// \brief Get the mapped x-coordinate
      ///////////////////////////////////////////////////////////////
      /// \param x: reference to mapped eye x-coordinate which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_x( real_t & x ) const {
         assert(m_parent);
         m_parent->lock_mutex( );

         if( m_mapper ) {
            x = m_mapper->map_x( m_x );
         } else {
            x = m_x;
         }

         m_parent->unlock_mutex( );
      }

      /// \brief Get the raw x-coordinate
      ///////////////////////////////////////////////////////////////
      /// \return the raw x-coordinate of the eye
      ///////////////////////////////////////////////////////////////
      inline real_t get_raw_x() const {
         real_t x;
         get_raw_x( x );
         return x;
      }

      /// \brief Get the x-coordinate
      ///////////////////////////////////////////////////////////////
      /// \param x: reference to raw eye x-coordinate which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_raw_x( real_t & x ) const {
         assert( m_parent );
         m_parent->lock_mutex( );
         x = m_x;
         m_parent->unlock_mutex( );
      }

      /// \brief Get mapped the y-coordinate
      ///////////////////////////////////////////////////////////////
      /// \return the mapped y-coordinate of the eye
      ///////////////////////////////////////////////////////////////
      inline real_t get_y( void ) const {
         real_t y;
         get_y( y );
         return y;
      }

      /// \brief Get the mapped y-coordinate
      ///////////////////////////////////////////////////////////////
      /// \param y: reference to mapped eye y-coordinate which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_y( real_t & y ) const {
         assert( m_parent );
         m_parent->lock_mutex( );

         if( m_mapper ) {
            y = m_mapper->map_y( m_y );
         } else {
            y = m_y;
         }

         m_parent->unlock_mutex( );
      }

      /// \brief Get the raw y-coordinate
      ///////////////////////////////////////////////////////////////
      /// \return the raw y-coordinate of the eye
      ///////////////////////////////////////////////////////////////
      inline real_t get_raw_y( void ) const {
         real_t y;
         get_raw_y( y );
         return y;
      }

      /// \brief Get the raw y-coordinate
      ///////////////////////////////////////////////////////////////
      /// \param y: reference to raw eye y-coordinate which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_raw_y( real_t & y ) const {
         assert( m_parent );
         m_parent->lock_mutex( );
         y = m_y;
         m_parent->unlock_mutex( );
      }

      /// \brief Get the diameter 
      ///////////////////////////////////////////////////////////////
      /// \return the diameter of the eye
      ///////////////////////////////////////////////////////////////
      inline real_t get_diameter( void ) const {
         real_t diameter;
         get_diameter( diameter );
         return diameter;
      }

      /// \brief Get the diameter 
      ///////////////////////////////////////////////////////////////
      /// \param diameter: reference to eye diameter which will be filled
      /// \return nothing
      ///////////////////////////////////////////////////////////////
      inline void get_diameter( real_t & diameter ) const {
         assert( m_parent );
         m_parent->lock_mutex( );
         diameter = m_diameter;
         m_parent->unlock_mutex( );
      }

    private:
      /// \brief Disable the Copy Constructor
      eye( const eye & );

      /// the parent iscan object
      iscan * m_parent;

      /// the mapper used to convert iscan data
      mapper * m_mapper;

      /// whether the class is respnsible for the deletion of the mapper
      bool m_responsible;

      /// the state of the eye: RESET, CALIBRATE, RUN
      state m_state;

      /// the x-coordinate of the eye
      real_t m_x;

      /// the y-coordinate of the eye
      real_t m_y;

      /// the diameter of the eye
      real_t m_diameter;
   };

   class update_callback 
   {
    public:
      virtual ~update_callback( void ) { }

      virtual void operator()( iscan *tracker ) = 0;
   };

   /// \brief Default Constructor
   iscan( void );

   /// \brief Conversion Constructor
   iscan( const std::string & device, mode m = BINOCULAR,
      dominance d = NO_DOMINANCE );

   /// \brief Destructor
   ~iscan( void );

   /// \brief Open the device in the given mode
   bool open( const std::string & device, mode m = BINOCULAR,
      dominance d = NO_DOMINANCE );

   /// \brief Set the dominance for the eyes
   inline void set_dominance( dominance d ) { 
      m_dominance = d;
   }

   /// \brief Get the dominance for the eyes
   inline dominance get_dominance( void ) { 
      return m_dominance;
   }

   /// \brief Close the device
   bool close( void );

   /// \brief Start the device
   bool start( void );

   /// \brief Stop the device
   bool stop( void );

   inline void set_update_callback( update_callback* callback ) {
      lock_mutex( );
      m_update_callback = callback;
      unlock_mutex( );
   }

   /// \brief Get the specified eye
   ///////////////////////////////////////////////////////////////
   /// \param num: the index into the eye array
   /// \return the specified eye
   ///////////////////////////////////////////////////////////////
   inline eye & get_eye( unsigned int num ) {
      assert( num < ISCAN_MAX_EYES );
      return m_eyes[ num ]; 
   }

   /// \brief Get the left eye
   ///////////////////////////////////////////////////////////////
   /// \return the left eye
   ///////////////////////////////////////////////////////////////
   inline eye & get_left_eye( void ) { 
      return get_eye( ISCAN_LEFT_EYE ); 
   }

   /// \brief Get the right eye
   ///////////////////////////////////////////////////////////////
   /// \return the right eye
   ///////////////////////////////////////////////////////////////
   inline eye & get_right_eye( void ) { 
      return get_eye(ISCAN_RIGHT_EYE); 
   }

 private:

   /// \brief Disable the Copy Constructor
   iscan( const iscan & );

   /// \brief Disable the Assignment Operator
   iscan & operator=( const iscan & );

   /// \brief Thread function to update iscan object.
   friend void * get_iscan_data( void * data );

   /// \brief Array of possible eyes
   eye m_eyes[ ISCAN_MAX_EYES ];

   /// \brief The number of eyes as specified by mode
   unsigned int m_num_eyes;

   /// \brief The dominance of the eyes
   dominance m_dominance;

   /// \brief The descriptor to read data from
   int m_device;

   /// \brief Whether the eye tracker is active or not
   bool m_active;

   /// \brief Thread handle
   pthread_t m_thread;

   /// \brief Mutex to protect critical sections
   mutable pthread_mutex_t m_mutex;

   /// \brief Saved IO settings for the device
   termios m_saved_settings;

   /// \brief Functor to be called upon the arrival of new data
   update_callback * m_update_callback;

   /// \brief Start the thread
   void create_thread();

   /// \brief Lock the mutex
   ///////////////////////////////////////////////////////////////
   /// \return nothing
   ///////////////////////////////////////////////////////////////
   inline void lock_mutex( void ) const { 
      pthread_mutex_lock( &m_mutex ); 
   }

   /// \brief Unlock the mutex
   ///////////////////////////////////////////////////////////////
   /// \return nothing
   ///////////////////////////////////////////////////////////////
   inline void unlock_mutex( void ) const { 
      pthread_mutex_unlock(&m_mutex); 
   }

   /// \brief Flushes data received, but not read, from the device
   ///////////////////////////////////////////////////////////////
   /// \return nothing
   ///////////////////////////////////////////////////////////////
   inline void clear_device( void ) { 
      tcflush( m_device, TCIFLUSH ); 
   }

   /// \brief Read the specified number of bytes into the buffer
   ///////////////////////////////////////////////////////////////
   /// \param buffer: the buffer to insert read data into
   /// \param size: the number of bytes to read
   /// \return nothing
   ///////////////////////////////////////////////////////////////
   bool read( byte * buffer, unsigned int size );

   /// \brief Get the next message from the device
   bool update( void );
};

/// \brief Print the eye status
std::ostream & operator<<( std::ostream & out, const iscan::eye & eye );

#endif
