//FILE:         ppm.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>

#include <iostream>
#include <fstream>
#include <string>
#include "ppm.h"
#include "image.h"
#include "gl_utils.h"

///////////////////////////////////////////////////////////////////////////////
static rgb_image*
ppm_ascii_load_rgb( std::istream& stream )
{
   //local vars
   std::string line;
   unsigned int rows, cols;
   unsigned int max, e, g, b;
   rgb_image::pixel_t color;

   //get rest of the magic number line
   getline( stream, line );

   //next line can be a comment
   while( stream.peek( ) == '#' ) {
      //this line is a comment
      getline( stream, line );
   }

   //read in width
   if( !(stream >> cols) ) return NULL;

   //read in height
   if( !(stream >> rows) ) return NULL;

   //read in max value
   if( !(stream >> max) ) return NULL;
   if( max == 0.0 ) return NULL;

   //create the image
   rgb_image *img = new rgb_image;
   img->resize( rows, cols );

   //read in each pixel
   for( unsigned int r = 0; r < rows; ++r ) {
      for( unsigned int c = 0; c < cols; ++c ) {
         if( !(stream >> e >> g >> b) ) {
            //either not enough data, or corrupt data
            delete img;
            return NULL;
         }

         //set pixel
         color[ 0 ] = static_cast<unsigned char>( e );
         color[ 1 ] = static_cast<unsigned char>( g );
         color[ 2 ] = static_cast<unsigned char>( b );
         img->set( r, c, color );
      }
   }

   //all done
   return img;
}



///////////////////////////////////////////////////////////////////////////////
static rgb_image*
ppm_binary_load_rgb( std::istream& stream )
{
   //local vars
   std::string line;
   unsigned int rows, cols;
   unsigned int max; 
   char e, g, b;
   rgb_image::pixel_t color;

   //get rest of the magic number line
   getline( stream, line );

   //next line can be a comment
   while( stream.peek( ) == '#' ) {
      //this line is a comment
      getline( stream, line );
   }

   //read in width
   if( !(stream >> cols) ) return NULL;

   //read in height
   if( !(stream >> rows) ) return NULL;

   //read in max value
   if( !(stream >> max) ) return NULL;
   if( max == 0 ) return NULL;

   //read in rest of max value line
   getline( stream, line );

   //create the image
   rgb_image *img = new rgb_image;
   img->resize( rows, cols );

   //read in each pixel
   for( unsigned int r = 0; r < rows; ++r ) {
      for( unsigned int c = 0; c < cols; ++c ) {
         if( !(stream.read( &e, 1 ) &&
            stream.read( &g, 1 ) &&
            stream.read( &b, 1 )) ) {
            //either not enough data, or corrupt data
            delete img;
            return NULL;
         }

         //set pixel
         color[ 0 ] = static_cast<unsigned char>( e );
         color[ 1 ] = static_cast<unsigned char>( g );
         color[ 2 ] = static_cast<unsigned char>( b );
         img->set( r, c, color );
      }
   }

   //all done
   return img;
}



///////////////////////////////////////////////////////////////////////////////
static grey_image*
ppm_ascii_load_grey( std::istream& stream )
{
   //local vars
   std::string line;
   unsigned int rows, cols;
   unsigned int max, grey;
   grey_image::pixel_t color;

   //get rest of the magic number line
   getline( stream, line );

   //next line can be a comment
   while( stream.peek( ) == '#' ) {
      //this line is a comment
      getline( stream, line );
   }

   //read in width
   if( !(stream >> cols) ) return NULL;

   //read in height
   if( !(stream >> rows) ) return NULL;

   //read in max value
   if( !(stream >> max) ) return NULL;
   if( max == 0.0 ) return NULL;

   //create the image
   grey_image *img = new grey_image;
   img->resize( rows, cols );

   //read in each pixel
   for( unsigned int r = 0; r < rows; ++r ) {
      for( unsigned int c = 0; c < cols; ++c ) {
         if( !(stream >> grey) ) {
            //either not enough data, or corrupt data
            delete img;
            return NULL;
         }

         //set pixel
         color[ 0 ] = static_cast<unsigned char>( grey );
         img->set( r, c, color );
      }
   }

   //all done
   return img;
}



///////////////////////////////////////////////////////////////////////////////
static grey_image*
ppm_binary_load_grey( std::istream& stream )
{
   //local vars
   std::string line;
   unsigned int rows, cols;
   unsigned int max; 
   char grey;
   grey_image::pixel_t color;

   //get rest of the magic number line
   getline( stream, line );

   //next line can be a comment
   while( stream.peek( ) == '#' ) {
      //this line is a comment
      getline( stream, line );
   }

   //read in width
   if( !(stream >> cols) ) return NULL;

   //read in height
   if( !(stream >> rows) ) return NULL;

   //read in max value
   if( !(stream >> max) ) return NULL;
   if( max == 0 ) return NULL;

   //read in rest of max value line
   getline( stream, line );

   //create the image
   grey_image *img = new grey_image;
   img->resize( rows, cols );

   //read in each pixel
   for( unsigned int r = 0; r < rows; ++r ) {
      for( unsigned int c = 0; c < cols; ++c ) {
         if( !(stream.read( &grey, 1 )) ) {
            //either not enough data, or corrupt data
            delete img;
            return NULL;
         }

         //set pixel
         color[ 0 ] = static_cast<unsigned char>( grey );
         img->set( r, c, color );
      }
   }

   //all done
   return img;
}


///////////////////////////////////////////////////////////////////////////////
static rgb_image*
load_rgb( std::istream& stream )
{
   //local vars
   char c;

   //first byte must be a 'P'
   stream.read( &c, 1 );
   if( c != 'P' ) return NULL;

   //next byte is a number
   stream.read( &c, 1 );
   if( c == '2' ) {
      std::cerr << "error: unsupported PNM type" << std::endl;
      return NULL;
   } else if( c == '3' ) {
      return ppm_ascii_load_rgb( stream );
   } else if( c == '5' ) {
      std::cerr << "error: unsupported PNM type" << std::endl;
      return NULL;
   } else if( c == '6' ) {
      return ppm_binary_load_rgb( stream );
   } else {
      return NULL;
   }
}



///////////////////////////////////////////////////////////////////////////////
static grey_image*
load_grey( std::istream& stream )
{
   //local vars
   char c;

   //first byte must be a 'P'
   stream.read( &c, 1 );
   if( c != 'P' ) return NULL;

   //next byte is a number
   stream.read( &c, 1 );
   if( c == '2' ) {
      return ppm_ascii_load_grey( stream );
      return NULL;
   } else if( c == '3' ) {
      std::cerr << "error: unsupported PNM type" << std::endl;
      return NULL;
   } else if( c == '5' ) {
      return ppm_binary_load_grey( stream );
   } else if( c == '6' ) {
      std::cerr << "error: unsupported PNM type" << std::endl;
      return NULL;
   } else {
      return NULL;
   }
}



///////////////////////////////////////////////////////////////////////////////
rgb_image*
ppm_load_rgb( const std::string& filename )
{
   //open file
   std::ifstream handle( filename.c_str( ) );
   if( !handle ) {
      //could not open file
      return NULL;
   }

   //load the file
   rgb_image* img = load_rgb( handle );

   //close the file
   handle.close( );

   //return the image
   return img;
}



///////////////////////////////////////////////////////////////////////////////
grey_image*
ppm_load_grey( const std::string& filename )
{
   //open file
   std::ifstream handle( filename.c_str( ) );
   if( !handle ) {
      //could not open file
      return NULL;
   }

   //load the file
   grey_image* img = load_grey( handle );

   //close the file
   handle.close( );

   //return the image
   return img;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
ppm_load_rgb_gl( const std::string& filename )
{
   rgb_image *img = ppm_load_rgb( filename );
   if( !img ) return gl_texture( );

   img->flip_horizontal( );

   gl_texture handle = gl_load( img );
   delete img;;
   return handle;
}



///////////////////////////////////////////////////////////////////////////////
gl_texture
ppm_load_grey_gl( const std::string& filename )
{
   grey_image *img = ppm_load_grey( filename );
   if( !img ) return gl_texture( );

   img->flip_horizontal( );

   gl_texture handle = gl_load( img );
   delete img;;
   return handle;
}



///////////////////////////////////////////////////////////////////////////////
bool
ppm_ascii_write( const std::string& filename, const rgb_image *img )
{
   assert( img );

   std::ofstream file( filename.c_str( ) );
   if( !file ) {
      std::cerr << "error: " << filename 
         << ": could not open file for writing" << std::endl;
      return false;
   }

   //write header
   file << "P3\n"
      << "#uber ppm output filter\n" 
      << img->w << " " << img->h << "\n"
      << "255\n"
      << std::endl;

   //write data
   unsigned char e, g, b;
   for( unsigned int r = 0; r < img->h; ++r ) {
      for( unsigned int c = 0; c < img->w; ++c ) {
         img->get( r, c, e, g, b );
         file << static_cast<int>( e ) << " "
            << static_cast<int>( g ) << " "
            << static_cast<int>( b ) << " ";
      }
   }
   file << std::endl;

   //success
   file.close( );
   return true;
}
