//FILE:         xine_dump.cpp
//AUTHOR:       Nathan Cournia <nathan@cournia.com>
//TODO:         I should probably put any references to m_finished as a
//              critical section.

#include <iostream>        //for debugging only
#include "xine_dump.h"
#include "xine/video_out_rgb.h"
#include "mathematics.h"

extern "C" 
{
///////////////////////////////////////////////////////////////////////////////
void xine_rgbout_callback( unsigned int width, unsigned int height, 
                           void *frame, void *data )
{
   xine_dump *dump = static_cast<xine_dump*>( data );
   dump->call_callback( width, height, frame );
}



///////////////////////////////////////////////////////////////////////////////
void xine_event_callback( void *data, const xine_event_t *event )
{
   if( XINE_EVENT_UI_PLAYBACK_FINISHED == event->type ) {
      xine_dump *dump = static_cast<xine_dump*>( data );
      dump->set_finished( true );
   }
}
} //end of extern "C"



///////////////////////////////////////////////////////////////////////////////
xine_dump::xine_dump( void ):
   m_xine( 0 ), m_stream( 0 ), m_vo_driver( 0 ), 
   m_ao_driver( 0 ), m_event_queue( 0 ),
   m_callback( 0 ), m_finished( false )
{
   //no-op
}



///////////////////////////////////////////////////////////////////////////////
xine_dump::~xine_dump( void )
{
   if( is_playing( ) ) {
      stop( );
   }

   if( m_stream ) {
      xine_close( m_stream );
   }

   if( m_vo_driver ) {
      xine_close_video_driver( m_xine, m_vo_driver );
   }

   if( m_ao_driver ) {
      xine_close_audio_driver( m_xine, m_ao_driver );
   }

   xine_exit( m_xine );
}



///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::init( void )
{
   //alloc xine memory
   m_xine = xine_new( );
   //look at setup_config

   //load xine config file
   std::string config_file = getenv( "HOME" );
   if( !config_file.empty( ) ) {
      config_file += "/.gcd/xine_config";
      xine_config_load( m_xine, config_file.c_str( ) );
   }

   //initialize xine
   xine_init( m_xine );

   rgbout_visual_info_t visual;
   visual.format = PX_BGR24;
   visual.levels = PXLEVEL_ALL;
   visual.callback = xine_rgbout_callback;
   visual.callback_data = this;

   /*
   const char ** p = (const char**)xine_list_video_output_plugins( m_xine );
   while( *p ) {
      std::cerr << "plug: " << *p << std::endl;
      ++p;
   }
   */

   //open video driver?
   std::cerr << "loading video driver" << std::endl;
   m_vo_driver = xine_open_video_driver( m_xine, "RGBout", 
      XINE_VISUAL_TYPE_RGBOUT, (void*)(&visual) );
   if( !m_vo_driver ) {
      set_error( "could not load video driver" );
      return false;
   }
   std::cerr << "loading video driver done" << std::endl;

   m_ao_driver = xine_open_audio_driver( m_xine, NULL, NULL );
   if( !m_ao_driver ) {
      std::cerr << "warning: could not load audio driver" << std::endl;
   }

   /*
   //we only want 5 video buffers
   xine_cfg_entry_t entry;
   config_lookup( "video.num_buffers", 5, &entry );
   entry.num_value = 5; //overwrite what was in the config file
   xine_config_update_entry( m_xine, &entry );
   */

   //create a xine stream
   m_stream = xine_stream_new( m_xine, m_ao_driver, m_vo_driver );

   //get pointer to event queue
   m_event_queue = xine_event_new_queue( m_stream );
   xine_event_create_listener_thread( m_event_queue, 
                                      xine_event_callback, this );

   clear_error( );
   return true;
}



///////////////////////////////////////////////////////////////////////////////
void
xine_dump::config_lookup( const char *id, int val, xine_cfg_entry_t *entry )
{
   memset( entry, 0, sizeof( xine_cfg_entry_t ) );
   if( !xine_config_lookup_entry( m_xine, id, entry ) ) {
      xine_config_register_num( m_xine, id, val, 0, NULL, 10, NULL, NULL );
      xine_config_lookup_entry( m_xine, id, entry );
   }
}




///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::open( const std::string& filename )
{
   //do we need to initialize xine?
   if( !m_xine ) {
      init( );
   }

   //was the init successful
   if( !m_xine ) {
      set_error( "could not initialize the xine engine" );
      return false;
   }

   //open the file
   if( !xine_open( m_stream, filename.c_str( ) ) ) {
      set_errorf( "%s: could not open file", filename.c_str( ) );
      return false;
   }

   //does this file have a video stream?
   if( 0 == xine_get_stream_info( m_stream, 
      XINE_STREAM_INFO_HAS_VIDEO ) ) {
      set_error( "file does not contain a video stream" );
      return false;
   }

   //can we handle the video
   if( 0 == xine_get_stream_info( m_stream, 
       XINE_STREAM_INFO_VIDEO_HANDLED ) ) {
      set_error( "unrecognized video stream" );
      return false;
   }

   //success
   clear_error( );
   return true;
}



///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::play( void )
{
   if( error( ) ) {
      return false;
   }

   if( !xine_play( m_stream, 0, 0 ) ) {
      set_error( "could not play stream" );
      return false;
   }

   m_finished = false;

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
void
xine_dump::stop( void )
{
   if( m_stream ) {
      xine_stop( m_stream );
   }
}



///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::is_playing( void )
{
   if( error( ) ) {
      return false;
   }

   return (XINE_STATUS_PLAY == xine_get_status( m_stream ) &&
      XINE_SPEED_NORMAL == xine_get_param( m_stream, XINE_PARAM_SPEED));
}



///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::is_finished( void )
{
   return m_finished;
}



///////////////////////////////////////////////////////////////////////////////
void
xine_dump::set_finished( bool x )
{
   m_finished = x;
}



///////////////////////////////////////////////////////////////////////////////
void
xine_dump::set_callback( xine_dump_callback *cb )
{
   m_callback = cb;
}



///////////////////////////////////////////////////////////////////////////////
bool
xine_dump::skip( int seconds )
{
   int perc, ms, len;
   if( !xine_get_pos_length( m_stream, &perc, &ms, &len ) ) {
      set_error( "could not get stream time info" );
      return false;
   }

   ms += seconds * 1000;
   ms = clamp( ms, 0, len );
   if( !xine_play( m_stream, 0, ms ) ) {
      set_error( "play: could not skip in stream" );
      return false;
   }

   //success
   return true;
}



///////////////////////////////////////////////////////////////////////////////
void
xine_dump::call_callback( unsigned int width, unsigned int height,
                          void *pixels )
{
   if( m_callback ) {
      (*m_callback)( width, height,
                     static_cast<unsigned char*>( pixels ) );
   }
}
